# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
# SPDX-License-Identifier: MIT


"""
sudo pip3 install adafruit-blinka
sudo pip3 install Jetson.GPIO
sudo pip3 install adafruit-circuitpython-neopixel-spi
"""

# https://docs.circuitpython.org/projects/neopixel/en/latest/api.html#neopixel.GRB
# class neopixel.NeoPixel(pin: Pin, n: int, *, bpp: int = 3, brightness: float = 1.0, auto_write: bool = True, pixel_order: str = None)[source]¶
# A sequence of neoself.pixels.

# Parameters:
# pin (Pin) – The pin to output neopixel data on.
# n (int) – The number of neoself.pixels in the chain
# bpp (int) – Bytes per pixel. 3 for RGB and 4 for RGBW self.pixels.
# brightness (float) – Brightness of the self.pixels between 0.0 and 1.0 where 1.0 is full brightness
# auto_write (bool) – True if the neoself.pixels should immediately change when set. If False, show must be called explicitly.
# pixel_order (str) – Set the pixel color channel order. GRBW is set by default.

import atexit
import logging
from enum import Enum

import board
import neopixel_spi as neopixel
from mev_monitor.colors import color_table

log = logging.getLogger(__name__)


Effect = Enum("Effect", ["static", "supercar", "blink"])


class Led:
    _simulate = False

    def __init__(self, config):
        self.config = config
        self.current_led = 0
        self.forward = True
        self.counter = 0
        self.effect: Effect = Effect.blink

        atexit.register(self.off)

        if self.config.leds > 0:
            spi = board.SPI()

            self.pixels = neopixel.NeoPixel_SPI(
                spi,
                config.leds,
                brightness=1.0,
                pixel_order=neopixel.GRB,
                auto_write=False,
            )

            log.info("led initialized with %d leds", self.config.leds)
        else:
            log.info("led simulated")
            self._simulate = True

        self.set("red")

    def set(self, color="red", brightness=1.0, effect=Effect.static):
        if self._simulate:
            log.info(
                "simulation: set color=%s(%x) effect=%s",
                color,
                color_table[color],
                effect,
            )
        else:
            self.color = color_table[color]
            self.effect = effect

    def off(self):
        self.set("black")

    def blink_once(self, led=0, color="blue"):
        if self.config.led_inverted:
            led = 2 if led==0 else 0
        self.pixels[led] = color_table[color]
        self.pixels.show()

    def update(self):
        self.counter += 1
        if self._simulate:
            return
        if self.effect == Effect.supercar:
            self._effect_supercar()
        elif self.effect == Effect.blink:
            self._effect_blink()
        else:
            self._effect_static()

    def _effect_blink(self):
        for i in range(self.config.leds):
            self.pixels[i] = 0 if self.counter % 2 else self.color
        self.pixels.show()

    def _effect_static(self):
        for i in range(self.config.leds):
            self.pixels[i] = self.color
        self.pixels.show()

    def _effect_supercar(self):
        self.pixels[self.current_led] = 0  # spento
        if self.forward:
            self.current_led += 1
            if self.current_led >= self.config.leds:
                self.current_led = self.config.leds - 1
                self.forward = False
        else:
            self.current_led -= 1
            if self.current_led < 0:
                self.current_led = 0
                self.forward = True
        self.pixels[self.current_led] = self.color
        self.pixels.show()


if __name__ == "__main__":
    import time

    from mev_monitor.config import get_config

    config = get_config()

    led = Led(config)

    led.set("red")
    # time.sleep(1)
    # led.setColor("green")
    # time.sleep(1)
    # led.setColor("orange")

    while True:
        led.update()
        time.sleep(0.2)
