import asyncio
import logging
import re
import subprocess
import sys
import time
from typing import Dict, List

import aiohttp
from autobahn.asyncio.wamp import ApplicationRunner, ApplicationSession
from autobahn.wamp.types import ComponentConfig
from mev_common.models import DeviceStatus
from pydantic import validate_call
from mev_common.models import Package, PACKAGES

from .led import Effect, Led

log = logging.getLogger(__name__)


class Interface(ApplicationSession):
    def __init__(self, config):
        ApplicationSession.__init__(self, ComponentConfig(config.realm))
        self.settings = config
        self.led = Led(config)
        self.device_status: DeviceStatus = None
        self._watchdog_brain = time.time()
        self.packages: List[Package] = []

        runner = ApplicationRunner(config.ws, config.realm, extra={})
        asyncio.get_event_loop().run_until_complete(runner.run(self, start_loop=False))

    async def onJoin(self, details):
        log.info("wamp session attached to realm %s", self.settings.realm)
        self.subscribe(self.on_device_status, "device.status")
        self.subscribe(self.on_upgrade, "device.upgrade.start")
        self.subscribe(self.on_plate, "plate.new")

        asyncio.create_task(self.set_info())

        #if self.settings.upgrade:
        #    asyncio.create_task(self.on_upgrade())

        asyncio.create_task(self.led_update())
        asyncio.create_task(self.watchdog_brain())
        if self.settings.push_frequency:
            asyncio.create_task(self.status_push())

    def on_plate(self, plate=None):
        log.debug("plate.new: quarter=%s", plate["quarter"])
        quarter = plate["quarter"]
        self.led.blink_once(led=2 if quarter==0 else 0)



    @validate_call
    async def on_device_status(self, status: DeviceStatus):
        self.device_status = status
        log.debug("device_status: %s", self.device_status)
        self._watchdog_brain = time.time()

        if self.device_status.yolodeep_status.up:
            if self.device_status.gps_status.satellites <= 0:
                self.led.set("blue")
            elif self.device_status.gps_status.satellites < 3:
                self.led.set("orange")
            else:
                self.led.set("green", effect=Effect.supercar)
        else:
            self.led.set("red", effect=Effect.blink)

    async def led_update(self, frequency=0.2):
        """loop refresh"""

        while True:
            self.led.update()
            await asyncio.sleep(frequency)

    async def watchdog_brain(self, timeout=6):
        """check brain status"""

        while True:
            if time.time() - self._watchdog_brain > timeout:
                self.led.set("red", effect=Effect.blink)
            await asyncio.sleep(timeout)

    async def status_push(self):
        """post status on remote service"""
        push_frequency = 3
        while True:
            if self.device_status:
                async with aiohttp.ClientSession(
                    timeout=aiohttp.ClientTimeout(total=self.settings.push_timeout)
                ) as session:
                    status = self.device_status.dict()
                    log.debug("push to %s: %s", self.settings.push_url, status)
                    try:
                        async with session.post(
                            self.settings.push_url, json=status
                        ) as resp:
                            payload = await resp.json()
                            log.debug("push response: %s", payload)
                            await self.status_check(payload)
                        push_frequency = self.settings.push_frequency
                    except Exception:
                        push_frequency = 3
                        log.warning(
                            "cannot push status to %s, retry in %d seconds",
                            self.settings.push_url,
                            push_frequency,
                        )
            await asyncio.sleep(push_frequency)

    async def status_check(self, payload: Dict):
        if self.settings.simulate_upgrade:
            version = ""
            for split in ("==", ">=", "<="):
                try:
                    package, version = self.settings.simulate_upgrade.split(split)
                    break
                except Exception:
                    pass
            await self.install(package, version)

        if "parsedBody" in payload:  # il payload arriva da echo test
            payload = payload["parsedBody"]
            payload["versions"] = {}
            # payload["versions"]["mev-frontend"] = "1.0.0" # simulate
        if "versions" in payload:
            started = time.time()
            for package in payload["versions"]:
                self.install(package, payload["versions"][package])
            self.publish("device.upgrade.end", elapsed=time.time() - started)

    def install(self, package: str, version: str, dry_run: bool = False):
        if version > "0.0.0":
            log.info(
                "install package %s==%s dry-run=%s",
                package,
                version,
                dry_run or self.settings.dry_run,
            )
            self.publish(
                "device.upgrade", state="running", package=package, version=version
            )
            try:
                if not (dry_run or self.settings.dry_run):
                    if version:
                        subprocess.check_call(
                            [
                                sys.executable,
                                "-m",
                                "pip",
                                "install",
                                f"{package}=={version}",
                            ]
                        )
                    else:
                        subprocess.check_call(
                            [sys.executable, "-m", "pip", "install", "-U", package]
                        )
                self.publish(
                    "device.upgrade", state="success", package=package, version=version
                )
                return True
            except Exception:
                log.error("failed to upgrade %s==%s", package, version)
                self.publish(
                    "device.upgrade", state="error", package=package, version=version
                )
            return False

    async def set_info(self):
        log.info("set_info() ...")
        loop = asyncio.get_running_loop()
        self.packages = await loop.run_in_executor(None, self.get_packages_info)
        log.info("packages: %s", self.packages)
        self.publish("device.packages", packages = [p.dict() for p in self.packages])

    async def on_upgrade(self, packages=[], dry_run=False):
        """
        packages è utilizzato solo per i test
        """
        #await self.set_info()
        package_to_restart = []
        for package in self.packages:
            # se sto facendo dei test
            if packages and package.name not in packages:
                continue 
            if package.latest_version and package.version != package.latest_version:
                if self.install(package.name, package.latest_version, dry_run=dry_run):
                    package_to_restart.append(package.name) 
        await self.restart(package_to_restart)
        await self.set_info()
        return package_to_restart

    def get_packages_info(self) -> List[Package]:
        """upgrade packages to latest version"""
        packages = []
        for package in PACKAGES:
            try:
                process = subprocess.run(
                    [sys.executable, "-m", "pip", "index", "versions", package],
                    stdout=subprocess.PIPE,
                )
                stdout = process.stdout.decode()
                try:
                    current_version = re.findall(".*INSTALLED:\\s*(.*)\n.*", stdout)[0]
                    latest_version = re.findall(".*LATEST:\\s*(.*)\n.*", stdout)[0]
                except Exception:
                    log.warning("cannot get latest version for package %s", package)
                    continue

                log.info(
                    "info for package %s: version=%s latest=%s",
                    package,
                    current_version,
                    latest_version,
                )
                packages.append(Package(name=package, version=current_version, latest_version=latest_version))
            except Exception as e:
                log.error("get_packages_info %s", str(e))
        return packages

    async def restart(self, packages):
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._restart_blocking, packages)

    def _restart_blocking(self, packages):
        if not self.settings.dry_run:
            subprocess.check_call(["systemctl", "--user", "daemon-reload"])
        for package in packages:
            if package in ("mev-monitor", "mev-brain"):
                continue
            if package == "mev-fake":
                service = "fake-api"
            else:
                service = package.split("-")[1]
            self._restart_service(service)
        if "mev-brain" in packages:
            self._restart_service("brain")
        if "mev-monitor" in packages:  # ultimo a essere chiamato
            self._restart_service("monitor")

    def _restart_service(self, service):
        log.info("restart %s service", service)
        if self.settings.dry_run:
            return
        try:
            subprocess.check_call(["systemctl", "--user", "restart", service])
        except Exception:
            log.error("failed to restart %s", service)

    def onDisconnect(self):
        log.warning("wamp transport lost, shutting down")
        asyncio.get_event_loop().stop()


def run():
    from .config import get_config

    config = get_config()

    Interface(config)
    try:
        asyncio.get_event_loop().run_forever()
    except KeyboardInterrupt:
        pass  # interface.close()
    finally:
        log.info("exit from app")


if __name__ == "__main__":
    run()
